==> ?obj=sdir%2F1.20%2Fsdir-source.zip: sdir-1.20/display.c <==
/*
 * DISPLAY.C
 *
 * Colorful, sorted and optionally rich directory enumeration
 * for Windows.
 *
 * This module implements low level routines to display formatted text to the
 * console or to files, in unicode or ansi as desired, pausing if too much
 * output has occurred, etc.
 *
 * Copyright (c) 2014 Malcolm J. Smith
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

#include "sdir.h"

//
//  Display support
//

BOOL
SdirWriteStringToFile (
    __in HANDLE hFile,
    __in_ecount(count) LPCTSTR str,
    __in DWORD count
    )
{
    DWORD written;

#ifdef UNICODE
    if (!Opts->OutputIsUnicode) {
        CHAR AsciiStackBuffer[64 + 1];
        LPSTR StringBuffer;
    
        if (count < sizeof(AsciiStackBuffer)) {
    
            WideCharToMultiByte(CP_ACP,
                                0,
                                str,
                                count,
                                AsciiStackBuffer,
                                sizeof(AsciiStackBuffer),
                                NULL,
                                NULL);
        
            return WriteFile(hFile, AsciiStackBuffer, count, &written, NULL);
    
        } else {
            BOOL ret;
    
            StringBuffer = HeapAlloc(GetProcessHeap(), 0, count + 1);
    
            //
            //  We'd normally display an error here, but we're probably here
            //  because we're trying to display an error - that's where large
            //  strings can arrive here.
            //
    
            if (StringBuffer == NULL) {
                return FALSE;
            }
    
            WideCharToMultiByte(CP_ACP,
                                0,
                                str,
                                count,
                                StringBuffer,
                                count + 1,
                                NULL,
                                NULL);
        
            ret = WriteFile(hFile, StringBuffer, count, &written, NULL);
    
            HeapFree(GetProcessHeap(), 0, StringBuffer);
    
            return ret;
        }
    } else {
        return WriteFile(hFile, str, count * sizeof(TCHAR), &written, NULL);
    }
#else
    return WriteFile(hFile, str, count * sizeof(TCHAR), &written, NULL);
#endif
}


BOOL
SdirWrite (
    __in_ecount(count) PSDIR_FMTCHAR str,
    __in DWORD count
    )
{
    HANDLE hConsole = GetStdHandle(STD_OUTPUT_HANDLE);
    DWORD written;
    DWORD i, j;

    WORD CacheAttr;
    TCHAR CharCache[64];

    //
    //  We firstly load as much as we can into a stack buffer, then write it
    //  out.  Those syscalls are expensive - run in a slow VM to see.  We
    //  have to write out the buffer either when the next character has 
    //  different color (so we can't use the same WriteConsole call) or when
    //  the stack buffer is full.
    //

    j = 0;
    if (Opts->OutputIsConsole) {
    
        for (i = 0; i < count; i++) { 
    
            CacheAttr = str[i].Attr;
            CharCache[j] = str[i].Char;
    
            j++;
    
            if (( i + 1 < count && str[i + 1].Attr != CacheAttr) || (j >= sizeof(CharCache)/sizeof(CharCache[0]))) {

                SetConsoleTextAttribute(hConsole, CacheAttr);
                WriteConsole(hConsole, CharCache, j, &written, NULL);
                j = 0;
            }
        }

    } else {

        for (i = 0; i < count; i++) { 
    
            CharCache[j] = str[i].Char;
            j++;
    
            if (j >= sizeof(CharCache)/sizeof(CharCache[0])) {

                SdirWriteStringToFile(hConsole, CharCache, j);
                j = 0;
            }
        }
    }

    //
    //  If we have anything left, flush it now.
    //

    if (j > 0) {

        if (Opts->OutputIsConsole) {

            SetConsoleTextAttribute(hConsole, CacheAttr);
            WriteConsole(hConsole, CharCache, j, &written, NULL);
        } else {
            SdirWriteStringToFile(hConsole, CharCache, j);
        }
    }

    return TRUE;
}

DWORD SdirWriteStringLinesDisplayed = 0;

BOOL
SdirWriteStringWithAttribute (
    __in LPCTSTR str,
    __in WORD DefaultAttribute
    )
{
    HANDLE hConsole = GetStdHandle(STD_OUTPUT_HANDLE);
    DWORD written;
    DWORD LinesInBuffer = 0;
    DWORD TCharsInBuffer = 0;
    BOOLEAN HitLineLimit = FALSE;

    if (Opts->OutputIsConsole) {

        while (str[TCharsInBuffer] != '\0') {
            while (str[TCharsInBuffer] != '\0') {
                if (str[TCharsInBuffer++] == '\n') {
                    LinesInBuffer++;
                    if (SdirWriteStringLinesDisplayed + LinesInBuffer >= Opts->ConsoleHeight - 1) {
                        HitLineLimit = TRUE;
                        break;
                    }
                }
            }

            SdirWriteStringLinesDisplayed += LinesInBuffer;

            SetConsoleTextAttribute(hConsole, DefaultAttribute);
            WriteConsole(hConsole, str, TCharsInBuffer, &written, NULL);

            LinesInBuffer = 0;
            str += TCharsInBuffer;
            TCharsInBuffer = 0;

            //
            //  This is screwy because we call it and it calls back into us.
            //  It works because we've just reset lines displayed to zero
            //  and we "know" we have enough space to display it.  But once
            //  we lose the hard coded value above, we need to guarantee to
            //  display at least two lines (one data, one press any key) to
            //  make progress.
            //

            if (HitLineLimit) {
                SdirWriteStringLinesDisplayed = 0;
                if (Opts->EnablePause && !SdirPressAnyKey()) {
                    return FALSE;
                }
                HitLineLimit = FALSE;
            }
        }

    } else {

        SdirWriteStringToFile(hConsole, str, (DWORD)_tcslen(str));
    }

    return TRUE;
}

BOOL
SdirPasteStrAndPad (
    __out_ecount(padsize) PSDIR_FMTCHAR str,
    __in_ecount(count) LPTSTR src,
    __in WORD attr,
    __in DWORD count,
    __in DWORD padsize
    )
{
    ULONG i;

    for (i = 0; i < count && i < padsize; i++) {
        str[i].Char = src[i];
        str[i].Attr = attr;
    }

    for (; i < padsize; i++) {
        str[i].Char = ' ';
        str[i].Attr = SDIR_DEFAULT_COLOR;
    }

    return TRUE;
}


BOOL
SdirDisplayError (
    __in LONG ErrorCode,
    __in LPCTSTR Prefix
    )
{
    LPTSTR OutputBuffer;

    if (FormatMessage(
            FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
            NULL,
            ErrorCode,
            MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
            (LPTSTR)&OutputBuffer,
            0,
            NULL)) {

        if (Prefix) {
            SdirWriteString(Prefix);
            SdirWriteString(_T(": "));
        }
        SdirWriteString(OutputBuffer);
        LocalFree((PVOID)OutputBuffer);
    }
    return TRUE;
}

BOOL
SdirPressAnyKey ()
{
    HANDLE hConsole = GetStdHandle(STD_INPUT_HANDLE);
    INPUT_RECORD InputBuffer;
    DWORD NumRead;

    SdirWriteString(_T("Press any key to continue..."));

    //
    //  Loop throwing away events until we get a key pressed
    //

    while (ReadConsoleInput(hConsole, &InputBuffer, 1, &NumRead) &&
           NumRead > 0 &&
           (InputBuffer.EventType != KEY_EVENT ||
            !InputBuffer.Event.KeyEvent.bKeyDown ||
            InputBuffer.Event.KeyEvent.uChar.UnicodeChar == '\0'));

    if (InputBuffer.Event.KeyEvent.uChar.UnicodeChar == 'q' ||
        InputBuffer.Event.KeyEvent.uChar.UnicodeChar == 'Q') {

        return FALSE;
    }

    //
    //  Keep throwing away any remaining events so we're not left
    //  with stale events next time we run
    //

    while (PeekConsoleInput(hConsole, &InputBuffer, 1, &NumRead) &&
           NumRead > 0) {

        ReadConsoleInput(hConsole, &InputBuffer, 1, &NumRead);
    }

    SdirWriteString(_T("\n"));
    SdirWriteStringLinesDisplayed = 0;
    return TRUE;
}

==> ?obj=sdir%2F1.21%2Fsdir-source.zip: sdir-1.21/display.c <==
/*
 * DISPLAY.C
 *
 * Colorful, sorted and optionally rich directory enumeration
 * for Windows.
 *
 * This module implements low level routines to display formatted text to the
 * console or to files, in unicode or ansi as desired, pausing if too much
 * output has occurred, etc.
 *
 * Copyright (c) 2014 Malcolm J. Smith
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

#include "sdir.h"

//
//  Display support
//

BOOL
SdirWriteStringToFile (
    __in HANDLE hFile,
    __in_ecount(count) LPCTSTR str,
    __in DWORD count
    )
{
    DWORD written;

#ifdef UNICODE
    if (!Opts->OutputIsUnicode) {
        CHAR AsciiStackBuffer[64 + 1];
        LPSTR StringBuffer;
    
        if (count < sizeof(AsciiStackBuffer)) {
    
            WideCharToMultiByte(CP_ACP,
                                0,
                                str,
                                count,
                                AsciiStackBuffer,
                                sizeof(AsciiStackBuffer),
                                NULL,
                                NULL);
        
            return WriteFile(hFile, AsciiStackBuffer, count, &written, NULL);
    
        } else {
            BOOL ret;
    
            StringBuffer = HeapAlloc(GetProcessHeap(), 0, count + 1);
    
            //
            //  We'd normally display an error here, but we're probably here
            //  because we're trying to display an error - that's where large
            //  strings can arrive here.
            //
    
            if (StringBuffer == NULL) {
                return FALSE;
            }
    
            WideCharToMultiByte(CP_ACP,
                                0,
                                str,
                                count,
                                StringBuffer,
                                count + 1,
                                NULL,
                                NULL);
        
            ret = WriteFile(hFile, StringBuffer, count, &written, NULL);
    
            HeapFree(GetProcessHeap(), 0, StringBuffer);
    
            return ret;
        }
    } else {
        return WriteFile(hFile, str, count * sizeof(TCHAR), &written, NULL);
    }
#else
    return WriteFile(hFile, str, count * sizeof(TCHAR), &written, NULL);
#endif
}


BOOL
SdirWrite (
    __in_ecount(count) PSDIR_FMTCHAR str,
    __in DWORD count
    )
{
    HANDLE hConsole = GetStdHandle(STD_OUTPUT_HANDLE);
    DWORD written;
    DWORD i, j;

    WORD CacheAttr;
    TCHAR CharCache[64];

    //
    //  We firstly load as much as we can into a stack buffer, then write it
    //  out.  Those syscalls are expensive - run in a slow VM to see.  We
    //  have to write out the buffer either when the next character has 
    //  different color (so we can't use the same WriteConsole call) or when
    //  the stack buffer is full.
    //

    j = 0;
    if (Opts->OutputIsConsole) {
    
        for (i = 0; i < count; i++) { 
    
            CacheAttr = str[i].Attr;
            CharCache[j] = str[i].Char;
    
            j++;
    
            if (( i + 1 < count && str[i + 1].Attr != CacheAttr) || (j >= sizeof(CharCache)/sizeof(CharCache[0]))) {

                SetConsoleTextAttribute(hConsole, CacheAttr);
                WriteConsole(hConsole, CharCache, j, &written, NULL);
                j = 0;
            }
        }

    } else {

        for (i = 0; i < count; i++) { 
    
            CharCache[j] = str[i].Char;
            j++;
    
            if (j >= sizeof(CharCache)/sizeof(CharCache[0])) {

                SdirWriteStringToFile(hConsole, CharCache, j);
                j = 0;
            }
        }
    }

    //
    //  If we have anything left, flush it now.
    //

    if (j > 0) {

        if (Opts->OutputIsConsole) {

            SetConsoleTextAttribute(hConsole, CacheAttr);
            WriteConsole(hConsole, CharCache, j, &written, NULL);
        } else {
            SdirWriteStringToFile(hConsole, CharCache, j);
        }
    }

    return TRUE;
}

DWORD SdirWriteStringLinesDisplayed = 0;

BOOL
SdirWriteStringWithAttribute (
    __in LPCTSTR str,
    __in WORD DefaultAttribute
    )
{
    HANDLE hConsole = GetStdHandle(STD_OUTPUT_HANDLE);
    DWORD written;
    DWORD LinesInBuffer = 0;
    DWORD TCharsInBuffer = 0;
    BOOLEAN HitLineLimit = FALSE;

    if (Opts->OutputIsConsole) {

        while (str[TCharsInBuffer] != '\0') {
            while (str[TCharsInBuffer] != '\0') {
                if (str[TCharsInBuffer++] == '\n') {
                    LinesInBuffer++;
                    if (SdirWriteStringLinesDisplayed + LinesInBuffer >= Opts->ConsoleHeight - 1) {
                        HitLineLimit = TRUE;
                        break;
                    }
                }
            }

            SdirWriteStringLinesDisplayed += LinesInBuffer;

            SetConsoleTextAttribute(hConsole, DefaultAttribute);
            WriteConsole(hConsole, str, TCharsInBuffer, &written, NULL);

            LinesInBuffer = 0;
            str += TCharsInBuffer;
            TCharsInBuffer = 0;

            //
            //  This is screwy because we call it and it calls back into us.
            //  It works because we've just reset lines displayed to zero
            //  and we "know" we have enough space to display it.  But once
            //  we lose the hard coded value above, we need to guarantee to
            //  display at least two lines (one data, one press any key) to
            //  make progress.
            //

            if (HitLineLimit) {
                SdirWriteStringLinesDisplayed = 0;
                if (Opts->EnablePause && !SdirPressAnyKey()) {
                    return FALSE;
                }
                HitLineLimit = FALSE;
            }
        }

    } else {

        while (*str != '\0') {

            //
            //  Look for line break or end of string
            //

            for (TCharsInBuffer = 0; str[TCharsInBuffer] != '\0' && str[TCharsInBuffer] != '\n'; TCharsInBuffer++);

            //
            //  If we have data, ignoring the line break or end of string, write it
            //

            if (TCharsInBuffer > 0) {

                SdirWriteStringToFile(hConsole, str, TCharsInBuffer);
            }

            //
            //  If we have a line break, use whatever desired break characters instead
            //

            if (str[TCharsInBuffer] == '\n') {

                SdirWriteStringToFile(hConsole, Opts->NewLine, (DWORD)_tcslen(Opts->NewLine));
                TCharsInBuffer++;
            }

            //
            //  And advance to the next portion of string
            //

            str += TCharsInBuffer;
        }
    }

    return TRUE;
}

BOOL
SdirPasteStrAndPad (
    __out_ecount(padsize) PSDIR_FMTCHAR str,
    __in_ecount(count) LPTSTR src,
    __in WORD attr,
    __in DWORD count,
    __in DWORD padsize
    )
{
    ULONG i;

    for (i = 0; i < count && i < padsize; i++) {
        str[i].Char = src[i];
        str[i].Attr = attr;
    }

    for (; i < padsize; i++) {
        str[i].Char = ' ';
        str[i].Attr = SDIR_DEFAULT_COLOR;
    }

    return TRUE;
}


BOOL
SdirDisplayError (
    __in LONG ErrorCode,
    __in LPCTSTR Prefix
    )
{
    LPTSTR OutputBuffer;

    if (FormatMessage(
            FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
            NULL,
            ErrorCode,
            MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
            (LPTSTR)&OutputBuffer,
            0,
            NULL)) {

        if (Prefix) {
            SdirWriteString(Prefix);
            SdirWriteString(_T(": "));
        }
        SdirWriteString(OutputBuffer);
        LocalFree((PVOID)OutputBuffer);
    }
    return TRUE;
}

BOOL
SdirPressAnyKey ()
{
    HANDLE hConsole = GetStdHandle(STD_INPUT_HANDLE);
    INPUT_RECORD InputBuffer;
    DWORD NumRead;

    SdirWriteString(_T("Press any key to continue..."));

    //
    //  Loop throwing away events until we get a key pressed
    //

    while (ReadConsoleInput(hConsole, &InputBuffer, 1, &NumRead) &&
           NumRead > 0 &&
           (InputBuffer.EventType != KEY_EVENT ||
            !InputBuffer.Event.KeyEvent.bKeyDown ||
            InputBuffer.Event.KeyEvent.uChar.UnicodeChar == '\0'));

    if (InputBuffer.Event.KeyEvent.uChar.UnicodeChar == 'q' ||
        InputBuffer.Event.KeyEvent.uChar.UnicodeChar == 'Q') {

        return FALSE;
    }

    //
    //  Keep throwing away any remaining events so we're not left
    //  with stale events next time we run
    //

    while (PeekConsoleInput(hConsole, &InputBuffer, 1, &NumRead) &&
           NumRead > 0) {

        ReadConsoleInput(hConsole, &InputBuffer, 1, &NumRead);
    }

    SdirWriteString(_T("\n"));
    SdirWriteStringLinesDisplayed = 0;
    return TRUE;
}

// vim:sw=4:ts=4:et:
